#include <stdio.h>
#include <ctype.h>
#include <string.h>
#include "turing.h"

#define MAX_STR	100
#define TRAKA_BLANK_CHAR ' '

TuringAkcija Turing::UcitajAkciju(FILE *f) {
	TuringAkcija akcija;
	char ts[MAX_STR];

	for (int i = 0; i < numtraka; ++i) {
		fscanf(f, "%s", ts);
		if (strlen(ts) == 3)
			akcija.zapis[i] = ts[1];
	}

	fscanf(f, "%s", ts);
	for (char *t = ts; *t; ++t) *t = toupper(*t);
	akcija.pomak = ts;

	fscanf(f, "%s", ts);
	akcija.iduce = &stanje[ts];
	stanje[ts].naziv = &stanje.find(ts)->first;

	return akcija;
}

TuringUlaz Turing::UcitajUlaz(FILE *f) const {
	TuringUlaz ulaz;
	char ts[MAX_STR];

	for (int i = 0; i < numtraka; ++i) {
		fscanf(f, "%s", ts);
		if (strlen(ts) == 3)
			ulaz[i] = ts[1];
	}

	return ulaz;
}

void Turing::RegisterTraceHandler(const Handler &h) {
	trace_handler = h;
}

void Turing::UnregisterTraceHandler() {
	trace_handler = Handler();
}

void Turing::CallHandler(const Handler &h) {
	if (h.f) h.f(this, h.data);
}

const TuringStanje * Turing::SetTrenutno(const TuringStanje *novo) {
	const TuringStanje *proslo = trenutno;
	trenutno = novo;
	CallHandler(trace_handler);
	return proslo;
}

bool Turing::Ucitaj(const string & fn) {
	FILE *f;
	char ts[MAX_STR];
	string name;
	int ntraka;

	f = fopen(fn.c_str(), "rt");
	if (!f) return false;

	if (fscanf(f, "%d tracks", &ntraka) != 1) {
		fclose(f);
		return false;
	}

	pocetno = NULL;
	SetTrenutno(NULL);
	stanje.clear();
	NoveTrake(ntraka);

	while (fscanf(f, "%s ", ts) == 1) {
		TuringStanje &novo = stanje[ts];
		novo.definirano = true;
		novo.naziv = &stanje.find(ts)->first;
		if (!pocetno) pocetno = &novo;

		fscanf(f, "%s", ts);
		for (;;) {
			fscanf(f, "%s", ts);
			if (strcmp(ts, ":")) break;
			TuringUlaz u = UcitajUlaz(f);
			novo.akcija[u] = UcitajAkciju(f);
		}
	}
	fclose(f);

	return true;
}

void Turing::DumpMemory() const {
	for (int i = 0; i < numtraka; ++i) {
		printf("Traka #%d: ", i);
		Traka[i].Dump();
	}
}

void TuringTraka::Dump() const {
	printf("%s$ @%d\n", c_str(), trenutno);
}

void TuringStanje::Dump() const {
	printf("Naziv: %s\n", GetNaziv().c_str());
	printf("Definirano: %d\n", definirano);
	if (!definirano) return;

	for (map<const TuringUlaz, TuringAkcija>::const_iterator i = akcija.begin(); i != akcija.end(); ++i) {
		i->first.Dump();
		printf(":=\n");
		i->second.Dump();
	}
	printf("\n");
}

void TuringUlaz::Dump() const {
	printf("{ ");
	for (const_iterator i = begin(); i != end(); ++i)
		printf("'%c'@%d ", i->second, i->first);
	printf("}");
}

void TuringAkcija::Dump() const {
	printf("Zapisi: ");
	for (map<const int, char>::const_iterator i = zapis.begin(); i != zapis.end(); ++i) {
		printf("'%c'->%d ", i->second, i->first);
	}
	printf("\n");
	printf("Pomak: %s\n", pomak.c_str());
	printf("Iduce: %s\n", iduce ? iduce->GetNaziv().c_str() : "NULL");
}

char TuringTraka::Procitaj() const {
	return (*this)[trenutno];
}

char TuringTraka::operator [] (unsigned int pos) const {
	if (pos >= size()) return 0;
	return (*dynamic_cast<const string*>(this))[pos];
}

bool TuringUlaz::Provjeri(const TuringTraka *traka) const {
	for (const_iterator i = begin(); i != end(); ++i)
		if (traka[i->first].Procitaj() != i->second)
			return false;
	return true;
}

const TuringStanje * Turing::Izvrsi(const TuringStanje *st, int *err) {
	map<const TuringUlaz, TuringAkcija>::const_iterator a;

	if (!st->definirano) {
		printf("-- ne mogu izvrsiti nedefinirano stanje \"%s\"!\n",
			st->GetNaziv().c_str());
		if (err) *err = 3;
		return NULL;
	}

	const TuringAkcija *ak = NULL, *uvijekak = NULL;
	for (a = st->akcija.begin(); a != st->akcija.end(); ++a) {
		if (a->first.Provjeri(Traka)) {
			if (a->first == TuringUlaz()) {
				uvijekak = &a->second;
			} else {
				ak = &a->second;
				break;
			}
		}
	}

	if (!ak && uvijekak) ak = uvijekak;
	if (ak) {
		for (int i = 0; i < numtraka; ++i) {
			if (TuringTraka::DekodPomak(ak->pomak[i]) == -1 &&
			    Traka[i].GetPozicija() == 0) {
				if (err) *err = 4;
				else printf("-- ne moze lijevije po traci!\n");
				return NULL;
			}
		}

		for (map<const int, char>::const_iterator b = ak->zapis.begin(); b != ak->zapis.end(); ++b)
			Traka[b->first].Zapisi(b->second);

		for (int i = 0; i < numtraka; ++i)
			Traka[i].Pomak(ak->pomak[i]);

		if (!ak->iduce->definirano) {
			if (err) *err = 2;
			else printf("-- prijelaz u nedefinirano stanje \"%s\"!\n", ak->iduce->GetNaziv().c_str());
			return NULL;
		}
		if (*err) *err = 0;
		return ak->iduce;
	}

	if (err) *err = 1;
	else printf("-- nema definiranog prijelaza!\n");
	return NULL;
}

bool Turing::IsReady() const {
	return !!trenutno;
}

void Turing::Init() {
	SetTrenutno(pocetno);

	for (int i = 0; i < numtraka; ++i)
		Traka[i].Pozicioniraj(0);
}

int Turing::Trace() {
	int greska;

	if (!trenutno) return -1;

	SetTrenutno(Izvrsi(trenutno, &greska));

	return greska;
}

void Turing::Simuliraj() {
	unsigned int i = 1;

	Init();

	do {
		SetTrenutno(Izvrsi(trenutno));
		printf("Korak %d:\n", i++);
		DumpMemory();
	} while (trenutno);
}

const TuringStanje * Turing::FindStanje(const string &name) const {
	ListaStanja::const_iterator i = stanje.find(name);
	if (i != stanje.end())
		return &i->second;
	return NULL;
}

void TuringTraka::Zapisi(char c) {
	char prosli;
	if (trenutno >= size()) {
		*this += string(trenutno - size() + 1, TRAKA_BLANK_CHAR);
	}
	prosli = (*dynamic_cast<string*>(this))[trenutno];
	if (prosli != c) {
		(*dynamic_cast<string*>(this))[trenutno] = c;
		CallHandler(change_handler);
	}
}

void TuringTraka::erase() {
	if (size()) {
		string::erase();
		CallHandler(change_handler);
	}

	if (trenutno) {
		trenutno = 0;
		CallHandler(move_handler);
	}
}

void TuringTraka::RegisterChangeHandler(const Handler &h) {
	change_handler = h;
}

void TuringTraka::RegisterMoveHandler(const Handler &h) {
	move_handler = h;
}

void TuringTraka::CallHandler(const Handler &h) {
	if (h.f) h.f(this, h.data);
}

int TuringTraka::DekodPomak(char c) {
	switch (c) {
		case 'L':
		case 'l':
			return -1;
		case 'D':
		case 'd':
		case 'R':
		case 'r':
			return 1;
	}
	return 0;
}

void TuringTraka::Pomak(char c) {
	switch (DekodPomak(c)) {
		case -1:
			if (trenutno == 0) {
				TuringGreska("Ne moze lijevije po traci!");
				break;
			}
			--trenutno;
			CallHandler(move_handler);
			break;

		case 1:
			++trenutno;
			CallHandler(move_handler);
			break;
	}
}

void TuringTraka::Pozicioniraj(unsigned int p) {
	if (trenutno == p) return;
	
	trenutno = p;

	CallHandler(move_handler);
}
